@@ -17,9 +17,11 @@ from threading import Thread
 
 import pluggy
 import py
+import six
 import toml
 from packaging import requirements
 from packaging.utils import canonicalize_name
+from packaging.version import Version
 
 import tox
 from tox.constants import INFO
@@ -92,18 +94,21 @@ class Parser:
                 super(HelpFormatter, self).__init__(prog, max_help_position=35, width=190)
 
         self.argparser = argparse.ArgumentParser(
-            description="tox options", add_help=False, prog="tox", formatter_class=HelpFormatter,
+            description="tox options",
+            add_help=False,
+            prog="tox",
+            formatter_class=HelpFormatter,
         )
         self._testenv_attr = []
 
     def add_argument(self, *args, **kwargs):
-        """ add argument to command line parser.  This takes the
+        """add argument to command line parser.  This takes the
         same arguments that ``argparse.ArgumentParser.add_argument``.
         """
         return self.argparser.add_argument(*args, **kwargs)
 
     def add_testenv_attribute(self, name, type, help, default=None, postprocess=None):
-        """ add an ini-file variable for "testenv" section.
+        """add an ini-file variable for "testenv" section.
 
         Types are specified as strings like "bool", "line-list", "string", "argv", "path",
         "argvlist".
@@ -120,7 +125,7 @@ class Parser:
         self._testenv_attr.append(VenvAttribute(name, type, default, help, postprocess))
 
     def add_testenv_attribute_obj(self, obj):
-        """ add an ini-file variable as an object.
+        """add an ini-file variable as an object.
 
         This works as the ``add_testenv_attribute`` function but expects
         "name", "type", "help", and "postprocess" attributes on the object.
@@ -237,8 +242,8 @@ class PosargsOption:
 
 class InstallcmdOption:
     name = "install_command"
-    type = "argv"
-    default = "python -m pip install {opts} {packages}"
+    type = "argv_install_command"
+    default = r"python -m pip install \{opts\} \{packages\}"
     help = "install command for dependencies and package under test."
 
     def postprocess(self, testenv_config, value):
@@ -312,7 +317,8 @@ def propose_configs(cli_config_file):
             from_folder = py.path.local(cli_config_file)
         else:
             print(
-                "ERROR: {} is neither file or directory".format(cli_config_file), file=sys.stderr,
+                "ERROR: {} is neither file or directory".format(cli_config_file),
+                file=sys.stderr,
             )
             return
     for basename in INFO.CONFIG_CANDIDATES:
@@ -333,7 +339,11 @@ def parse_cli(args, pm):
         raise SystemExit(0)
     interpreters = Interpreters(hook=pm.hook)
     config = Config(
-        pluginmanager=pm, option=option, interpreters=interpreters, parser=parser, args=args,
+        pluginmanager=pm,
+        option=option,
+        interpreters=interpreters,
+        parser=parser,
+        args=args,
     )
     return config, option
 
@@ -382,7 +392,7 @@ class SetenvDict(object):
                 return os.environ.get(name, default)
             self._lookupstack.append(name)
             try:
-                self.resolved[name] = res = self.reader._replace(val)
+                self.resolved[name] = res = self.reader._replace(val, name="setenv")
             finally:
                 self._lookupstack.pop()
             return res
@@ -400,11 +410,26 @@ class SetenvDict(object):
         self.definitions[name] = value
         self.resolved[name] = value
 
+    def items(self):
+        return ((name, self[name]) for name in self.definitions)
+
+    def export(self):
+        # post-process items to avoid internal syntax/semantics
+        # such as {} being escaped using \{\}, suitable for use with
+        # os.environ .
+        return {
+            name: Replacer._unescape(value)
+            for name, value in self.items()
+            if value is not self._DUMMY
+        }
+
 
 @tox.hookimpl
 def tox_addoption(parser):
     parser.add_argument(
-        "--version", action="store_true", help="report version information to stdout.",
+        "--version",
+        action="store_true",
+        help="report version information to stdout.",
     )
     parser.add_argument("-h", "--help", action="store_true", help="show help about options")
     parser.add_argument(
@@ -434,7 +459,9 @@ def tox_addoption(parser):
         help="show list of all defined environments (with description if verbose)",
     )
     parser.add_argument(
-        "-c", dest="configfile", help="config file name or directory with 'tox.ini' file.",
+        "-c",
+        dest="configfile",
+        help="config file name or directory with 'tox.ini' file.",
     )
     parser.add_argument(
         "-e",
@@ -453,10 +480,14 @@ def tox_addoption(parser):
     )
     parser.add_argument("--notest", action="store_true", help="skip invoking test commands.")
     parser.add_argument(
-        "--sdistonly", action="store_true", help="only perform the sdist packaging activity.",
+        "--sdistonly",
+        action="store_true",
+        help="only perform the sdist packaging activity.",
     )
     parser.add_argument(
-        "--skip-pkg-install", action="store_true", help="skip package installation for this run",
+        "--skip-pkg-install",
+        action="store_true",
+        help="skip package installation for this run",
     )
     add_parallel_flags(parser)
     parser.add_argument(
@@ -493,7 +524,10 @@ def tox_addoption(parser):
         "(pip by default).",
     )
     parser.add_argument(
-        "-r", "--recreate", action="store_true", help="force recreation of virtual environments",
+        "-r",
+        "--recreate",
+        action="store_true",
+        help="force recreation of virtual environments",
     )
     parser.add_argument(
         "--result-json",
@@ -621,7 +655,9 @@ def tox_addoption(parser):
                     warnings.warn(
                         "conflicting basepython version (set {}, should be {}) for env '{}';"
                         "resolve conflict or set ignore_basepython_conflict".format(
-                            proposed_version, implied_version, testenv_config.envname,
+                            proposed_version,
+                            implied_version,
+                            testenv_config.envname,
                         ),
                     )
 
@@ -637,7 +673,7 @@ def tox_addoption(parser):
 
     def merge_description(testenv_config, value):
         """the reader by default joins generated description with new line,
-         replace new line with space"""
+        replace new line with space"""
         return value.replace("\n", " ")
 
     parser.add_testenv_attribute(
@@ -649,11 +685,17 @@ def tox_addoption(parser):
     )
 
     parser.add_testenv_attribute(
-        name="envtmpdir", type="path", default="{envdir}/tmp", help="venv temporary directory",
+        name="envtmpdir",
+        type="path",
+        default="{envdir}/tmp",
+        help="venv temporary directory",
     )
 
     parser.add_testenv_attribute(
-        name="envlogdir", type="path", default="{envdir}/log", help="venv log directory",
+        name="envlogdir",
+        type="path",
+        default="{envdir}/log",
+        help="venv log directory",
     )
 
     parser.add_testenv_attribute(
@@ -768,7 +810,9 @@ def tox_addoption(parser):
     )
 
     parser.add_testenv_attribute(
-        name="whitelist_externals", type="line-list", help="DEPRECATED: use allowlist_externals",
+        name="whitelist_externals",
+        type="line-list",
+        help="DEPRECATED: use allowlist_externals",
     )
 
     parser.add_testenv_attribute(
@@ -969,7 +1013,7 @@ class TestenvConfig:
         #: set of factors
         self.factors = factors
         self._reader = reader
-        self._missing_subs = []
+        self._missing_subs = {}
         """Holds substitutions that could not be resolved.
 
         Pre 2.8.1 missing substitutions crashed with a ConfigError although this would not be a
@@ -977,6 +1021,20 @@ class TestenvConfig:
         check later when the testenv is actually run and crash only then.
         """
 
+    # Python 3 only, as __getattribute__ is ignored for old-style types on Python 2
+    def __getattribute__(self, name):
+        rv = object.__getattribute__(self, name)
+        if isinstance(rv, Exception):
+            raise rv
+        return rv
+
+    if six.PY2:
+
+        def __getattr__(self, name):
+            if name in self._missing_subs:
+                raise self._missing_subs[name]
+            raise AttributeError(name)
+
     def get_envbindir(self):
         """Path to directory where scripts/binaries reside."""
         is_bin = (
@@ -1089,7 +1147,10 @@ class ParseIni(object):
         context_name = getcontextname()
         if context_name == "jenkins":
             reader = SectionReader(
-                "tox:jenkins", self._cfg, prefix=prefix, fallbacksections=[fallbacksection],
+                "tox:jenkins",
+                self._cfg,
+                prefix=prefix,
+                fallbacksections=[fallbacksection],
             )
             dist_share_default = "{toxworkdir}/distshare"
         elif not context_name:
@@ -1237,7 +1298,7 @@ class ParseIni(object):
         requires_list = reader.getlist("requires")
         config.minversion = reader.getstring("minversion", None)
         config.provision_tox_env = name = reader.getstring("provision_tox_env", ".tox")
-        min_version = "tox >= {}".format(config.minversion or tox.__version__)
+        min_version = "tox >= {}".format(config.minversion or Version(tox.__version__).public)
         deps = self.ensure_requires_satisfied(config, requires_list, min_version)
         if config.run_provision:
             section_name = "testenv:{}".format(name)
@@ -1245,7 +1306,10 @@ class ParseIni(object):
                 self._cfg.sections[section_name] = {}
             self._cfg.sections[section_name]["description"] = "meta tox"
             env_config = self.make_envconfig(
-                name, "{}{}".format(testenvprefix, name), reader._subs, config,
+                name,
+                "{}{}".format(testenvprefix, name),
+                reader._subs,
+                config,
             )
             env_config.deps = deps
             config.envconfigs[config.provision_tox_env] = env_config
@@ -1300,7 +1364,10 @@ class ParseIni(object):
             self._cfg.sections[section_name]["sitepackages"] = "False"
             self._cfg.sections[section_name]["description"] = "isolated packaging environment"
             config.envconfigs[name] = self.make_envconfig(
-                name, "{}{}".format(testenvprefix, name), reader._subs, config,
+                name,
+                "{}{}".format(testenvprefix, name),
+                reader._subs,
+                config,
             )
 
     def _list_section_factors(self, section):
@@ -1334,13 +1401,17 @@ class ParseIni(object):
                     "dict_setenv",
                     "argv",
                     "argvlist",
+                    "argv_install_command",
                 ):
                     meth = getattr(reader, "get{}".format(atype))
                     res = meth(env_attr.name, env_attr.default, replace=replace)
                 elif atype == "basepython":
                     no_fallback = name in (config.provision_tox_env,)
                     res = reader.getstring(
-                        env_attr.name, env_attr.default, replace=replace, no_fallback=no_fallback,
+                        env_attr.name,
+                        env_attr.default,
+                        replace=replace,
+                        no_fallback=no_fallback,
                     )
                 elif atype == "space-separated-list":
                     res = reader.getlist(env_attr.name, sep=" ")
@@ -1354,9 +1425,10 @@ class ParseIni(object):
                 if env_attr.postprocess:
                     res = env_attr.postprocess(testenv_config=tc, value=res)
             except tox.exception.MissingSubstitution as e:
-                tc._missing_subs.append(e.name)
-                res = e.FLAG
-            setattr(tc, env_attr.name, res)
+                tc._missing_subs[env_attr.name] = res = e
+            # On Python 2, exceptions are handled in __getattr__
+            if not six.PY2 or not isinstance(res, Exception):
+                setattr(tc, env_attr.name, res)
             if atype in ("path", "string", "basepython"):
                 reader.addsubstitutions(**{env_attr.name: res})
         return tc
@@ -1420,7 +1492,7 @@ class ParseIni(object):
         The parser will see it as two different sections: [testenv:py36-cov], [testenv:py37-cov]
 
         """
-        factor_re = re.compile(r"\{\s*([\w\s,]+)\s*\}")
+        factor_re = re.compile(r"\{\s*([\w\s,-]+)\s*\}")
         split_re = re.compile(r"\s*,\s*")
         to_remove = set()
         for section in list(config.sections):
@@ -1515,7 +1587,15 @@ is_section_substitution = re.compile(r"{\[[^{}\s]+\]\S+?}").match
 
 
 class SectionReader:
-    def __init__(self, section_name, cfgparser, fallbacksections=None, factors=(), prefix=None):
+    def __init__(
+        self,
+        section_name,
+        cfgparser,
+        fallbacksections=None,
+        factors=(),
+        prefix=None,
+        posargs="",
+    ):
         if prefix is None:
             self.section_name = section_name
         else:
@@ -1526,6 +1606,7 @@ class SectionReader:
         self._subs = {}
         self._subststack = []
         self._setenv = None
+        self.posargs = posargs
 
     def get_environ_value(self, name):
         if self._setenv is None:
@@ -1563,13 +1644,21 @@ class SectionReader:
         if value is None or not replace:
             return default or {}
 
-        d = {}
+        env_values = {}
         for line in value.split(sep):
             if line.strip():
-                name, rest = line.split("=", 1)
-                d[name.strip()] = rest.strip()
-
-        return d
+                if line.startswith("#"):  # comment lines are ignored
+                    pass
+                elif line.startswith("file|"):  # file markers contain paths to env files
+                    file_path = line[5:].strip()
+                    if os.path.exists(file_path):
+                        with open(file_path, "rt") as file_handler:
+                            content = file_handler.read()
+                        env_values.update(self._getdict(content, "", sep, replace))
+                else:
+                    name, value = line.split("=", 1)
+                    env_values[name.strip()] = value.strip()
+        return env_values
 
     def getfloat(self, name, default=None, replace=True):
         s = self.getstring(name, default, replace=replace)
@@ -1605,11 +1694,27 @@ class SectionReader:
 
     def getargvlist(self, name, default="", replace=True):
         s = self.getstring(name, default, replace=False)
-        return _ArgvlistReader.getargvlist(self, s, replace=replace)
+        return _ArgvlistReader.getargvlist(self, s, replace=replace, name=name)
 
     def getargv(self, name, default="", replace=True):
         return self.getargvlist(name, default, replace=replace)[0]
 
+    def getargv_install_command(self, name, default="", replace=True):
+        s = self.getstring(name, default, replace=False)
+        if not s:
+            # This occurs when factors are used, and a testenv doesnt have
+            # a factorised value for install_command, most commonly occurring
+            # if setting platform is also used.
+            # An empty value causes error install_command must contain '{packages}'.
+            s = default
+
+        if "{packages}" in s:
+            s = s.replace("{packages}", r"\{packages\}")
+        if "{opts}" in s:
+            s = s.replace("{opts}", r"\{opts\}")
+
+        return _ArgvlistReader.getargvlist(self, s, replace=replace, name=name)[0]
+
     def getstring(self, name, default=None, replace=True, crossonly=False, no_fallback=False):
         x = None
         sections = [self.section_name] + ([] if no_fallback else self.fallbacksections)
@@ -1634,6 +1739,17 @@ class SectionReader:
         x = self._replace_if_needed(x, name, replace, crossonly)
         return x
 
+    def getposargs(self, default=None):
+        if self.posargs:
+            posargs = self.posargs
+            if sys.platform.startswith("win"):
+                posargs_string = list2cmdline([x for x in posargs if x])
+            else:
+                posargs_string = " ".join([shlex_quote(x) for x in posargs if x])
+            return posargs_string
+        else:
+            return default or ""
+
     def _replace_if_needed(self, x, name, replace, crossonly):
         if replace and x and hasattr(x, "replace"):
             x = self._replace(x, name=name, crossonly=crossonly)
@@ -1660,6 +1776,7 @@ class SectionReader:
             return value
 
         section_name = section_name if section_name else self.section_name
+        assert name
         self._subststack.append((section_name, name))
         try:
             replaced = Replacer(self, crossonly=crossonly).do_replace(value)
@@ -1680,7 +1797,7 @@ class Replacer:
         (?<!\\)[{]
         (?:(?P<sub_type>[^[:{}]+):)?    # optional sub_type for special rules
         (?P<substitution_value>(?:\[[^,{}]*\])?[^:,{}]*)  # substitution key
-        (?::(?P<default_value>[^{}]*))?   # default value
+        (?::(?P<default_value>([^{}]|\\{|\\})*))?   # default value
         [}]
         """,
         re.VERBOSE,
@@ -1706,6 +1823,10 @@ class Replacer:
 
         return expanded
 
+    @staticmethod
+    def _unescape(s):
+        return s.replace("\\{", "{").replace("\\}", "}")
+
     def _replace_match(self, match):
         g = match.groupdict()
         sub_value = g["substitution_value"]
@@ -1716,40 +1837,47 @@ class Replacer:
             start, end = match.span()
             return match.string[start:end]
 
-        # special case: all empty values means ":" which is os.pathsep
-        if not any(g.values()):
+        full_match = match.group(0)
+        # ":" is swallowed by the regex, so the raw matched string is checked
+        if full_match.startswith("{:"):
+            if full_match != "{:}":
+                raise tox.exception.ConfigError(
+                    "Malformed substitution with prefix ':': {}".format(full_match),
+                )
+
             return os.pathsep
 
-        # special case: opts and packages. Leave {opts} and
-        # {packages} intact, they are replaced manually in
-        # _venv.VirtualEnv.run_install_command.
-        if sub_value in ("opts", "packages"):
-            return "{{{}}}".format(sub_value)
+        default_value = g["default_value"]
+        if sub_value == "posargs":
+            return self.reader.getposargs(default_value)
 
-        try:
-            sub_type = g["sub_type"]
-        except KeyError:
+        sub_type = g["sub_type"]
+        if not sub_type and not sub_value:
             raise tox.exception.ConfigError(
-                "Malformed substitution; no substitution type provided",
+                "Malformed substitution; no substitution type provided. "
+                "If you were using `{}` for `os.pathsep`, please use `{:}`.",
             )
 
+        if not sub_type and not default_value and sub_value == "/":
+            return os.sep
+
         if sub_type == "env":
-            return self._replace_env(match)
+            return self._replace_env(sub_value, default_value)
         if sub_type == "tty":
             if is_interactive():
                 return match.group("substitution_value")
             return match.group("default_value")
+        if sub_type == "posargs":
+            return self.reader.getposargs(sub_value)
         if sub_type is not None:
             raise tox.exception.ConfigError(
                 "No support for the {} substitution type".format(sub_type),
             )
-        return self._replace_substitution(match)
+        return self._replace_substitution(sub_value)
 
-    def _replace_env(self, match):
-        key = match.group("substitution_value")
+    def _replace_env(self, key, default):
         if not key:
             raise tox.exception.ConfigError("env: requires an environment variable name")
-        default = match.group("default_value")
         value = self.reader.get_environ_value(key)
         if value is not None:
             return value
@@ -1764,18 +1892,20 @@ class Replacer:
             cfg = self.reader._cfg
             if section in cfg and item in cfg[section]:
                 if (section, item) in self.reader._subststack:
-                    raise ValueError(
+                    raise tox.exception.SubstitutionStackError(
                         "{} already in {}".format((section, item), self.reader._subststack),
                     )
                 x = str(cfg[section][item])
                 return self.reader._replace(
-                    x, name=item, section_name=section, crossonly=self.crossonly,
+                    x,
+                    name=item,
+                    section_name=section,
+                    crossonly=self.crossonly,
                 )
 
         raise tox.exception.ConfigError("substitution key {!r} not found".format(key))
 
-    def _replace_substitution(self, match):
-        sub_key = match.group("substitution_value")
+    def _replace_substitution(self, sub_key):
         val = self.reader._subs.get(sub_key, None)
         if val is None:
             val = self._substitute_from_other_section(sub_key)
@@ -1790,7 +1920,7 @@ def is_interactive():
 
 class _ArgvlistReader:
     @classmethod
-    def getargvlist(cls, reader, value, replace=True):
+    def getargvlist(cls, reader, value, replace=True, name=None):
         """Parse ``commands`` argvlist multiline string.
 
         :param SectionReader reader: reader to be used.
@@ -1812,28 +1942,23 @@ class _ArgvlistReader:
             current_command += line
 
             if is_section_substitution(current_command):
-                replaced = reader._replace(current_command, crossonly=True)
-                commands.extend(cls.getargvlist(reader, replaced))
+                replaced = reader._replace(current_command, crossonly=True, name=name)
+                commands.extend(cls.getargvlist(reader, replaced, name=name))
             else:
-                commands.append(cls.processcommand(reader, current_command, replace))
+                commands.append(cls.processcommand(reader, current_command, replace, name=name))
             current_command = ""
         else:
             if current_command:
                 raise tox.exception.ConfigError(
                     "line-continuation ends nowhere while resolving for [{}] {}".format(
-                        reader.section_name, "commands",
+                        reader.section_name,
+                        "commands",
                     ),
                 )
         return commands
 
     @classmethod
-    def processcommand(cls, reader, command, replace=True):
-        posargs = getattr(reader, "posargs", "")
-        if sys.platform.startswith("win"):
-            posargs_string = list2cmdline([x for x in posargs if x])
-        else:
-            posargs_string = " ".join([shlex_quote(x) for x in posargs if x])
-
+    def processcommand(cls, reader, command, replace=True, name=None):
         # Iterate through each word of the command substituting as
         # appropriate to construct the new command string. This
         # string is then broken up into exec argv components using
@@ -1841,19 +1966,14 @@ class _ArgvlistReader:
         if replace:
             newcommand = ""
             for word in CommandParser(command).words():
-                if word == "{posargs}" or word == "[]":
-                    newcommand += posargs_string
+                if word == "[]":
+                    newcommand += reader.getposargs()
                     continue
-                elif word.startswith("{posargs:") and word.endswith("}"):
-                    if posargs:
-                        newcommand += posargs_string
-                        continue
-                    else:
-                        word = word[9:-1]
+
                 new_arg = ""
-                new_word = reader._replace(word)
-                new_word = reader._replace(new_word)
-                new_word = new_word.replace("\\{", "{").replace("\\}", "}")
+                new_word = reader._replace(word, name=name)
+                new_word = reader._replace(new_word, name=name)
+                new_word = Replacer._unescape(new_word)
                 new_arg += new_word
                 newcommand += new_arg
         else:
